import datetime as dt
import calendar

import pymssql
import sqlalchemy as sa
import numpy as np
import pandas as pd
from pathlib import Path


MSSQL_CONN = "mssql+pymssql://readdnds:reader%402021@121.37.138.1:14331/dnds"
MYSQL_CONN = "mysql+pymysql://reader:Reader#2020@121.37.138.1:13316/vision_product"
# MYSQL_CONN = "mysql+pymysql://rl_sync:rl_sync_2019@121.37.138.1:13306/vision"

RISK_STYLES = ['BETA',
               'MOMENTUM',
               'SIZE',
               'EARNYILD',
               'RESVOL',
               'GROWTH',
               'BTOP',
               'LEVERAGE',
               'LIQUIDTY',
               'SIZENL']

INDUSTRY_STYLES = [
    'Bank',
    'RealEstate',
    'Health',
    'Transportation',
    'Mining',
    'NonFerMetal',
    'HouseApp',
    'LeiService',
    'MachiEquip',
    'BuildDeco',
    'CommeTrade',
    'CONMAT',
    'Auto',
    'Textile',
    'FoodBever',
    'Electronics',
    'Computer',
    'LightIndus',
    'Utilities',
    'Telecom',
    'AgriForest',
    'CHEM',
    'Media',
    'IronSteel',
    'NonBankFinan',
    'ELECEQP',
    'AERODEF',
    'Conglomerates'
]


BOND_CLASSES = ["990301", "990302", "990303", "990304", "990305", "990306", "990399", "990501", "990502", "990205", "990402", "990603"]


def read_sql(query, source="mssql"):
    if source == "mssql":
        engine = sa.create_engine(MSSQL_CONN, encoding="utf8", connect_args={"charset": "GBK"})
    elif source == "mysql":
        engine = sa.create_engine(MYSQL_CONN)
    return pd.read_sql(query, con=engine)


# def nearest_report_date(date, freq="Q", method="forward"):
#     if len(date) == 10:
#         date = date.replace("-", "")

#     mm = int(date[4:6])
#     year = int(date[0:4])
#     last_year = year
    
#     if method == "forward":
#         add = 1
#     else:
#         add = 0
    
#     if freq == "Q":
#         q = (mm // 3 + add) * 3 - 3
#     else:
#         q = (mm // 6 + add) * 6 - 6
        
#     if q <= 0:
#         q += 12
#         last_year -= 1
#     last = calendar.monthrange(last_year, q)[1]
#     return dt.date(last_year, q, last).strftime("%Y%m%d")

def nearest_report_date(date, freq="Q"):
    if len(date) == 10:
        date = date.replace("-", "")

    mm = int(date[4:6])
    year = int(date[0:4])
    last_year = year

    if freq == "Q":
        q = ((mm - 1) // 3) * 3
    else:
        q = ((mm - 1) // 6) * 6

    if q == 0:
        q = 12
        last_year = year - 1
    last = calendar.monthrange(last_year, q)[1]
    return dt.date(last_year, q, last).strftime("%Y%m%d")


def fd_alive_funds(trade_dt, class_code):
    if class_code:
        query = f"""
                select security_code as SECURITYID, begin_date as BEGINDATE, end_date as ENDDATE, l2_name as CLASSCODE FROM wd_fd_classification
                WHERE
                    flag = 1 AND
                    begin_date <= '{_to_yyyy_mm_dd(trade_dt)}' AND
                    (end_date>='{_to_yyyy_mm_dd(trade_dt)}' or end_date = '19000101') AND
                    l1_name = '{class_code}'
                ORDER BY security_code;
                """ 
    else:
        query = f"""
                select security_code as SECURITYID, begin_date as BEGINDATE, end_date as ENDDATE, l2_name as CLASSCODE FROM wd_fd_classification
                WHERE
                    flag = 1 AND
                    begin_date <= '{_to_yyyy_mm_dd(trade_dt)}' AND
                    (end_date>='{_to_yyyy_mm_dd(trade_dt)}' or end_date = '19000101') 
                ORDER BY security_code;
                """
    return read_sql(query, "mysql")


def fd_basicinfo(security_ids=None, trade_dt=None):
    if not trade_dt:
        trade_dt = dt.datetime.today().strftime("%Y%m%d")
    else:
        trade_dt = trade_dt.replace("-", "")

    if isinstance(security_ids, str):
        sec_id_strs = security_ids
    elif not security_ids:
        sec_id_strs = None
    else:
        sec_id_strs = ",".join(["'" + s + "'" for s in security_ids])
        
    if sec_id_strs:
        query = f"""
        select SECODE as SECURITYID, FDNAME, SNAMECOMP, FSYMBOL, FDNATURE, INVESTSTYLE FROM TQ_FD_BASICINFO
        WHERE
            ISVALID = 1 AND
            SECODE in ({sec_id_strs}) AND
            (LIQUENDDATE >= '{trade_dt}' or LIQUENDDATE = '19000101')
        """
    else:
        query = f"""
        select SECODE as SECURITYID, FDNAME, SNAMECOMP, FSYMBOL, FDNATURE, INVESTSTYLE FROM TQ_FD_BASICINFO
        WHERE
            ISVALID = 1 AND
            (LIQUENDDATE >= '{trade_dt}' or LIQUENDDATE = '19000101')
        """
    return read_sql(query)


def fd_typeclass(security_ids, trade_dt=None):
    if not trade_dt:
        trade_dt = dt.datetime.today().strftime("%Y%m%d")
    else:
        trade_dt = trade_dt.replace("-", "")

    if isinstance(security_ids, str):
        sec_id_strs = security_ids
    else:
        sec_id_strs = ",".join(["'" + s + "'" for s in security_ids])

    query = f"""
    select TQ_FD_BASICINFO.SECODE as SECURITYID,
           TQ_FD_TYPECLASS.L1CODE,
           TQ_FD_TYPECLASS.L1NAME,
           TQ_FD_TYPECLASS.L2CODE,
           TQ_FD_TYPECLASS.L2NAME,
           TQ_FD_TYPECLASS.L3CODE,
           TQ_FD_TYPECLASS.L3NAME from TQ_FD_TYPECLASS join TQ_FD_BASICINFO on TQ_FD_BASICINFO.SECURITYID = TQ_FD_TYPECLASS.SECURITYID
    WHERE
        TQ_FD_TYPECLASS.ISVALID = 1 AND
        TQ_FD_BASICINFO.ISVALID = 1 AND
        TQ_FD_BASICINFO.SECODE in ({sec_id_strs}) AND
        (TQ_FD_TYPECLASS.ENDDATE >= '{trade_dt}' or TQ_FD_TYPECLASS.ENDDATE = '19000101')
    """
    return read_sql(query)


def fd_hshkiport(security_ids, report_dates_begin, report_dates_end=None):

    if isinstance(security_ids, str):
        sec_id_strs = security_ids
    else:
        sec_id_strs = ",".join(["'" + s + "'" for s in security_ids])
    
    if report_dates_end:
        query = f"""
        select SECODE as SECURITYID, INDCLASSCODE, INDUSTRYCODE, INDUSTRYNAME, REPORTDATE, MVALUE, ACCNETMKTCAP FROM TQ_FD_HSHKIPORT
        WHERE
            REPORTDATE >= '{report_dates_begin}' AND
            REPORTDATE <= '{report_dates_end}' AND
            ISVALID = 1 AND
            INDCLASSCODE = '2102' AND
            SECODE in ({sec_id_strs})
        """
    else:
        query = f"""
        select SECODE as SECURITYID, INDCLASSCODE, INDUSTRYCODE, INDUSTRYNAME, REPORTDATE, MVALUE, ACCNETMKTCAP FROM TQ_FD_HSHKIPORT
        WHERE
            REPORTDATE >= '{report_dates_begin}' AND
            ISVALID = 1 AND
            INDCLASSCODE = '2102' AND
            SECODE in ({sec_id_strs})
        """
    return read_sql(query).sort_values("SECURITYID").reset_index(drop=True)


def _to_yyyy_mm_dd(trade_dt):
    if len(trade_dt) == 10:
        return trade_dt
    else:
        return f"{trade_dt[0:4]}-{trade_dt[4:6]}-{trade_dt[6:8]}"
    
    
def _to_yyyymmdd(trade_dt):
    if len(trade_dt) == 8:
        return trade_dt
    else:
        return f"{trade_dt[0:4]}{trade_dt[5:7]}{trade_dt[8:10]}"
    

def fd_assetportfolio(security_ids, report_dates_begin, report_dates_end=None):
    # 获取相关组合情况
    if isinstance(security_ids, str):
        sec_id_strs = security_ids
    else:
        sec_id_strs = ",".join(["'" + s + "'" for s in security_ids])
    if report_dates_end:
        query = f"""
        SELECT security_code as SECURITYID, report_date as REPORTDATE, bd_rto as BDRTO, conv_bd_rto as CONVBDRTO, equity_inve_rto as EQUITYINVERTO, st_fin_bd_rto as STFINBDRTO, deposit_rto as DEPOSITRTO FROM fd_portfolio
        WHERE
            report_date >= '{report_dates_begin}' AND
            report_date <= '{report_dates_end}' AND
            is_valid = 1 AND
            flag = 1 AND
            publish_date != 19000101 AND
            security_code in ({sec_id_strs})
        """
    else:
        query = f"""
        SELECT security_code as SECURITYID, report_date as REPORTDATE, bd_rto as BDRTO, conv_bd_rto as CONVBDRTO, equity_inve_rto as EQUITYINVERTO, st_fin_bd_rto as STFINBDRTO, deposit_rto as DEPOSITRTO FROM fd_portfolio
        WHERE
            report_date >= '{report_dates_begin}' AND
            is_valid = 1 AND
            flag = 1 AND
            publish_date != 19000101 AND
            security_code in ({sec_id_strs})
        """
    return read_sql(query, source="mysql").sort_values("SECURITYID").reset_index(drop=True)


def fd_qtfdnav(security_ids, trade_dt):
    if isinstance(security_ids, str):
        sec_id_strs = security_ids
    else:
        sec_id_strs = ",".join(["'" + s + "'" for s in security_ids])

    query = f"""
    SELECT SECODE as SECURITYID, UNITNAV, UNITACCNAV, REPAIRUNITNAV, NAVGRTD FROM TQ_FD_DERIVEDN
    WHERE
        ENDDATE = '{trade_dt}' AND
        ISVALID = 1 AND
        SECODE in ({sec_id_strs})
    """
    return read_sql(query).sort_values("SECURITYID").reset_index(drop=True)





def _join_ids(security_ids):
    if isinstance(security_ids, str):
        sec_id_strs = "'" + security_ids + "'"
    elif security_ids is None:
        sec_id_strs = None
    else:
        sec_id_strs = ",".join(["'" + s + "'" for s in security_ids])
    return sec_id_strs


def fd_skdetail(security_ids, trade_dt):
    p_id_strs = _join_ids(security_ids)

    query = f"""
    SELECT SECODE as pid, PUBLISHDATE, SKCODE as security_code, SKNAME, HOLDMKTCAP, HOLDAMT, NAVRTO FROM TQ_FD_SKDETAIL
    WHERE
        ENDDATE = '{_to_yyyymmdd(trade_dt)}' AND
        ISVALID = 1 AND
        SECODE in ({p_id_strs})
    """
    groups = read_sql(query).sort_values("security_code").groupby("pid")
    dfs = []
    for k, g in groups:
        # g = g[g.PUBLISHDATE == np.max(g.PUBLISHDATE)]
        total_value = g.HOLDMKTCAP.sum()
        g["weight"] = g.HOLDMKTCAP / total_value
        dfs.append(g)
    df = pd.concat(dfs)
    del df["PUBLISHDATE"]
    return df.reset_index(drop=True)


def risk_exposure(start_date, end_date, security_ids=None):
    sec_id_strs = _join_ids(security_ids)

    if sec_id_strs:
        query = f"""
        SELECT trade_date, security_code, {','.join(RISK_STYLES)}, {','.join(INDUSTRY_STYLES)} FROM risk_exposure
        WHERE 
            trade_date >= '{_to_yyyy_mm_dd(start_date)}' AND
            trade_date <= '{_to_yyyy_mm_dd(end_date)}' AND
            flag = 1 AND
            security_code IN ({sec_id_strs})
        """
    else:
        query = f"""
        SELECT trade_date, security_code, {','.join(RISK_STYLES)}, {','.join(INDUSTRY_STYLES)} FROM risk_exposure
        WHERE 
            trade_date >= '{_to_yyyy_mm_dd(start_date)}' AND
            trade_date <= '{_to_yyyy_mm_dd(end_date)}' AND
            flag = 1
        """
    return read_sql(query, source="mysql").sort_values(["trade_date", "security_code"]).reset_index(drop=True)


def risk_return(start_date, end_date):
    query = f"""
            SELECT trade_date, {','.join(RISK_STYLES)}, {','.join(INDUSTRY_STYLES)} FROM risk_return
            WHERE 
                trade_date >= '{_to_yyyy_mm_dd(start_date)}' AND
                trade_date <= '{_to_yyyy_mm_dd(end_date)}' AND
                flag = 1
            """
    return read_sql(query, source="mysql").sort_values(["trade_date"]).reset_index(drop=True)


def specific_return(start_date, end_date, security_ids=None):
    sec_id_strs = _join_ids(security_ids)

    if sec_id_strs:
        query = f"""
        SELECT trade_date, security_code,  secShortName, symbol, spret FROM specific_return
        WHERE 
          trade_date >= '{_to_yyyy_mm_dd(start_date)}' AND
          trade_date <= '{_to_yyyy_mm_dd(end_date)}' AND
          flag = 1 AND
            security_code IN ({sec_id_strs})
        """
    else:
        query = f"""
        SELECT trade_date, security_code,  secShortName, symbol, spret FROM specific_return
        WHERE 
          trade_date >= '{_to_yyyy_mm_dd(start_date)}' AND
          trade_date <= '{_to_yyyy_mm_dd(end_date)}' AND
          flag = 1
        """
    return read_sql(query, source="mysql").sort_values(["trade_date", "security_code"]).reset_index(drop=True)


def risk_cov(start_date, end_date, model="short"):
    query = f"""
        SELECT trade_date, FactorID, {','.join(RISK_STYLES)}, {','.join(INDUSTRY_STYLES)} FROM risk_cov_{model}
        WHERE 
            trade_date >= '{_to_yyyy_mm_dd(start_date)}' AND
            trade_date <= '{_to_yyyy_mm_dd(end_date)}' AND
            flag = 1
        """
    return read_sql(query, source="mysql").sort_values(["trade_date", "FactorID"]).reset_index(drop=True)


def special_risk(start_date, end_date, model="short", security_ids=None):
    sec_id_strs = _join_ids(security_ids)

    if sec_id_strs:
        query = f"""
        SELECT trade_date, security_code, secShortName, symbol, SRISK FROM specific_risk_{model}
        WHERE 
            trade_date >= '{_to_yyyy_mm_dd(start_date)}' AND
            trade_date <= '{_to_yyyy_mm_dd(end_date)}' AND
            flag = 1 AND
            security_code IN ({sec_id_strs})
        """
    else:
        query = f"""
        SELECT trade_date, security_code, secShortName, symbol, SRISK FROM specific_risk_{model}
        WHERE 
            trade_date >= '{_to_yyyy_mm_dd(start_date)}' AND
            trade_date <= '{_to_yyyy_mm_dd(end_date)}' AND
            flag = 1
        """
    return read_sql(query, source="mysql").sort_values(["trade_date", "security_code"]).reset_index(drop=True)


def index_component(index_code, start_date, end_date):
    query = f"""
            SELECT index_component.trade_date, index_component.security_code, `index`.sname, `index`.weighing / 100.0 AS weight FROM `index`
            JOIN index_component ON
              `index`.trade_date = index_component.trade_date AND
              `index`.isymbol = index_component.isymbol AND
              `index`.symbol = index_component.symbol
            WHERE
              `index`.trade_date >= '{_to_yyyy_mm_dd(start_date)}' AND
              `index`.trade_date <= '{_to_yyyy_mm_dd(end_date)}' AND
              `index`.flag = 1 AND
              index_component.flag = 1 AND
              index_component.isecurity_code = '{index_code}';
              """
    return read_sql(query, source="mysql").sort_values(["trade_date", "security_code"]).reset_index(drop=True)


def sw_industry_daily(start_date, end_date, level=1, security_ids=None):
    sec_id_strs = _join_ids(security_ids)
    if sec_id_strs:
        query = f"""
        SELECT trade_date, security_code, industry_code{level} as industry_code, industry_name{level} as industry_name FROM sw_industry_daily_2021
        WHERE
          trade_date >= '{_to_yyyy_mm_dd(start_date)}' AND
          trade_date <= '{_to_yyyy_mm_dd(end_date)}' AND
          flag = 1 AND
          security_code IN ({sec_id_strs})
        """
    else:
        query = f"""
        SELECT trade_date, security_code, industry_code{level} as industry_code, industry_name{level} as industry_name FROM sw_industry_daily_2021
        WHERE
          trade_date >= '{_to_yyyy_mm_dd(start_date)}' AND
          trade_date <= '{_to_yyyy_mm_dd(end_date)}' AND
          flag = 1
        """
    return read_sql(query, source="mysql").sort_values(["trade_date", "security_code"]).reset_index(drop=True)


def stk_return(start_date, end_date, security_ids=None):
    sec_id_strs = _join_ids(security_ids)
    if sec_id_strs:
        query = f"""
        SELECT security_code, trade_date, `close` * lat_factor AS lat_price FROM stk_daily_price_pro
        WHERE
              trade_date in ('{_to_yyyy_mm_dd(start_date)}', '{_to_yyyy_mm_dd(end_date)}') AND
              security_code IN ({sec_id_strs}) AND
              flag = 1
        """
    else:
        query = f"""
        SELECT security_code, trade_date, `close` * lat_factor AS lat_price  FROM stk_daily_price_pro
        WHERE
              trade_date in ('{_to_yyyy_mm_dd(start_date)}', '{_to_yyyy_mm_dd(end_date)}') AND
              flag = 1
        """
        
    df  = read_sql(query, source="mysql").sort_values("security_code")
    df["trade_date"] = df["trade_date"].astype(str)
    df = df.pivot(index="security_code", columns="trade_date", values="lat_price").dropna()
    df.rename(columns={f'{_to_yyyy_mm_dd(start_date)}': "start_date", f'{_to_yyyy_mm_dd(end_date)}': "end_date"}, inplace=True)
    df["chg_pct"] = df["end_date"] / df["start_date"] - 1.0
    return df.reset_index()


def fd_manager(manager_ids=None, manager_names=None, start_date=None, end_date=None):
    if manager_ids:
        manager_ids = _join_ids(manager_ids)
    elif manager_names:
        manager_names = _join_ids(manager_names)

    if manager_ids:
        query = f"""
        SELECT MANAGERCODE, MANAGERNAME, COMPCODE, COMPNAME, SECODE as SECURITYID, BEGINDATE, ENDDATE FROM TQ_FD_MGPERFORMANCE
        WHERE
            BEGINDATE <= '{_to_yyyymmdd(end_date)}' AND
            (ENDDATE >= '{_to_yyyymmdd(start_date)}' OR ENDDATE = '19000101') AND
            MANAGERCODE in ({manager_ids}) AND
            ISVALID = 1
        """
    elif manager_names:
        query = f"""
        SELECT MANAGERCODE, MANAGERNAME, COMPCODE, COMPNAME, SECODE as SECURITYID, BEGINDATE, ENDDATE FROM TQ_FD_MGPERFORMANCE
        WHERE
            BEGINDATE <= '{_to_yyyymmdd(end_date)}' AND
            (ENDDATE >= '{_to_yyyymmdd(start_date)}' OR ENDDATE = '19000101') AND
            MANAGERNAME in ({manager_names}) AND
            ISVALID = 1
        """
    else:
        query = f"""
        SELECT MANAGERCODE, MANAGERNAME, COMPCODE, COMPNAME, SECODE as SECURITYID, BEGINDATE, ENDDATE FROM TQ_FD_MGPERFORMANCE
        WHERE
            BEGINDATE <= '{_to_yyyymmdd(end_date)}' AND
            (ENDDATE >= '{_to_yyyymmdd(start_date)}' OR ENDDATE = '19000101') AND
            ISVALID = 1
        """
    return read_sql(query).sort_values("MANAGERCODE").reset_index(drop=True)


def fd_fund_mktv(security_ids, trade_dt):
    nav_df = fd_qtfdnav(security_ids, trade_dt)
    sec_id_strs = _join_ids(security_ids)

    query = f"""
    SELECT SECODE as SECURITYID, REPORTDATE, INIFDSHARE * 10000 as INIFDSHARE, ENDFDSHARE * 10000 as ENDFDSHARE FROM TQ_FD_SHARE
    WHERE
        REPORTDATE <= '{_to_yyyymmdd(trade_dt)}' AND
        SECODE IN ({sec_id_strs}) AND
        ISVALID = 1
    ORDER BY REPORTDATE
    """
    shares_df = read_sql(query).sort_values(["SECURITYID", "REPORTDATE"]).reset_index(drop=True)
    shares_df = shares_df.groupby("SECURITYID").last().reset_index()
    df = pd.merge(nav_df[["SECURITYID", "UNITNAV", "NAVGRTD"]], shares_df[["SECURITYID", "ENDFDSHARE"]], on="SECURITYID")
    df["MKTV"] = df["UNITNAV"] * df["ENDFDSHARE"]
    return df


def fd_rl_classification(security_ids, trade_dt, source="hb"):
    sec_id_strs = _join_ids(security_ids)
    if source == "rl":
        table = "fd_rl_classification"
        query = f"""
        SELECT security_id as SECURITYID,  rl_class, rl_class_name, rl_sub_class, rl_sub_class_name FROM {table}
        WHERE
            trade_date = '{_to_yyyymmdd(trade_dt)}' AND
            security_id in ({sec_id_strs}) AND
            flag = 1
        """
    elif source == "hb":
        table = "hb_fd_classification"
        query = f"""
        SELECT security_code as SECURITYID,  rl_class, rl_class_name, rl_sub_class, rl_sub_class_name FROM {table}
        WHERE
            security_code in ({sec_id_strs}) AND
            flag = 1
        """
    return read_sql(query, source="mysql").sort_values("SECURITYID").reset_index(drop=True)


def read_factors(security_ids, trade_dt, factors):
    sec_id_strs = _join_ids(security_ids)
    
    query = f"""
    SELECT security_code, trade_date, {','.join(factors)}  FROM factor_valuation_estimation
    WHERE
        trade_date = '{_to_yyyymmdd(trade_dt)}' AND
        security_code in ({sec_id_strs}) AND
        flag = 1
    """
    return read_sql(query, source="mysql").sort_values("security_code").reset_index(drop=True)


def index_rtn(security_ids, begin_date, end_date):
    sec_id_strs = _join_ids(security_ids)
    
    query = f"""
    SELECT security_code, min(trade_date) AS begin_date, max(trade_date) AS end_date, exp(sum(ln(1 + change_pct / 100.))) - 1.0 AS rtn FROM index_daily_price
    WHERE
        trade_date >= '{_to_yyyy_mm_dd(begin_date)}' AND
        trade_date <= '{_to_yyyy_mm_dd(end_date)}' AND
        security_code in ({sec_id_strs}) AND
        flag = 1
    GROUP BY
        security_code
    """
    df = read_sql(query, source="mysql").sort_values("security_code").reset_index(drop=True)
    return df
